/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam;

import dk.vajhoej.record.RecordException;

/**
 * Interface IsamSource represents an ISAM source.
 */
public interface IsamSource {
	/**
	 * Read record from ISAM source.
	 * @param t type to read
	 * @param k key of record to read
	 * @return object read
	 * @throws IsamException if problem with ISAM source
	 * @throws RecordException if problem converting data
	 */
	public <T,TK> T read(Class<T> t, Key<TK> k) throws IsamException, RecordException;
	/**
	 * Create record in ISAM source.
	 * @param o object to create
	 * @throws IsamException if problem with ISAM source
	 * @throws RecordException if problem converting data
	 */
	public <T> void create(T o) throws IsamException, RecordException;
    /**
     * Update record in ISAM source.
     * @param o object to update
     * @throws IsamException if problem with ISAM source
     * @throws RecordException if problem converting data
     */
    public <T> void update(T o) throws IsamException, RecordException;
    /**
     * Delete record from ISAM source.
     * @param t type to delete
     * @param k key of record to delete
     * @throws IsamException if problem with ISAM source
     * @throws RecordException if problem converting data
     */
    public <T,TK> void delete(Class<T> t, Key<TK> k) throws IsamException, RecordException;
    /**
     * Close ISAM source.
     * @throws IsamException if problem with ISAM source
     */
	public void close() throws IsamException;
	/**
	 * Find all records greater than or equal key.
	 * @param t type to find
	 * @param k key to find
	 * @return result
	 * @throws IsamException if problem with ISAM source
	 * @throws RecordException if problem converting data
	 */
	public <T,TK> IsamResult<T> readGE(Class<T> t, Key<TK> k) throws IsamException, RecordException;
	/**
	 * Read all records greater than key.
	 * @param t type to find
	 * @param k key to find
	 * @return result
	 * @throws IsamException if problem with ISAM source
	 * @throws RecordException if problem converting data
	 */
	public <T,TK> IsamResult<T> readGT(Class<T> t, Key<TK> k) throws IsamException, RecordException;
	/**
	 * Read all records less than or equal key.
     * Note that the performance of reading reverse may be very poor in some implementations.
	 * @param t type to find
	 * @param k key to find
	 * @return result
	 * @throws IsamException if problem with ISAM source
	 * @throws RecordException if problem converting data
	 */
	public <T,TK> IsamResult<T> readLE(Class<T> t, Key<TK> k) throws IsamException, RecordException;
	/**
	 * Read all records less than key.
     * Note that the performance of reading reverse may be very poor in some implementations.
	 * @param t type to find
	 * @param k key to find
	 * @return result
	 * @throws IsamException if problem with ISAM source
	 * @throws RecordException if problem converting data
	 */
	public <T,TK> IsamResult<T> readLT(Class<T> t, Key<TK> k) throws IsamException, RecordException;
}
