/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.emulated;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.logging.Level;
import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamResult;
import dk.vajhoej.isam.IsamSource;
import dk.vajhoej.isam.Key;
import dk.vajhoej.isam.KeyInfo;
import dk.vajhoej.isam.KeyInfoArrayCache;
import dk.vajhoej.record.FieldType;
import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.StructReader;
import dk.vajhoej.record.StructWriter;

/**
 * Class EmulatedIsamSource is emulating an ISAM source by using a relational database.
 */
public class EmulatedIsamSource implements IsamSource {
    private static Logger log = Logger.getLogger(EmulatedIsamSource.class.getName());
	private static final String PARAMS = "?,?,?,?,?,?,?,?,?,?,?";
	private Connection con;
	private String conurl;
	private String un;
	private String pw;
	/**
	 * Create instance of EmulatedIsamSource.
	 * @param driver JDBC driver class
	 * @param conurl JDBC connection URL
	 * @param un username used to connect
	 * @param pw password used to connect
	 * @throws IsamException if problem with ISAM source
	 */
	public EmulatedIsamSource(String driver, String conurl, String un, String pw) throws IsamException {
		try {
			Class.forName(driver);
            con = DriverManager.getConnection(conurl, un, pw);
            this.conurl = conurl;
            this.un = un;
            this.pw = pw;
            log.fine("Connected to " + conurl);
		} catch (ClassNotFoundException e) {
			throw new IsamException("Database driver not found", e);
		} catch (SQLException e) {
			throw new IsamException("Database problem", e);
		}
	}
	public <T, TK> T read(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("Read class " + t.getName() + " with key " + k);
		KeyInfo[] keys = KeyInfoArrayCache.analyze(t);
		PreparedStatement pstmt = null;
		ResultSet rs = null;
		try {
			pstmt = con.prepareStatement("SELECT recblob FROM " + t.getSimpleName() + " WHERE " + keys[k.getIx()].getField().getName() + " = ?");
			setKey(pstmt, keys, k);
			rs = pstmt.executeQuery();
			if(rs.next()) {
			    log.fine("Found");
			    byte[] res = rs.getBytes(1);
	            log.finest("Length: " + res.length);
			    if(log.isLoggable(Level.FINEST)) {
			        log.finest("Bytes:" + LogHelper.byteArrayToString(res));
			    }
				StructReader sr = new StructReader(res);
				return sr.read(t);
			} else {
			    log.fine("Not found");
				return null;
			}
		} catch (SQLException e) {
			throw new IsamException("Database problem", e);
		} finally {
			try {
				if(rs != null) rs.close();
				if(pstmt != null) pstmt.close();
			} catch (SQLException e) {
				// not much that can be done
			}
		}
	}
	public <T> void create(T o) throws IsamException, RecordException {
        log.fine("Create class " + o.getClass().getName() + " with value " + o);
		Class<?> t = o.getClass();
		KeyInfo[] keys = KeyInfoArrayCache.analyze(t);
		PreparedStatement pstmt = null;
		try {
			StructWriter sw = new StructWriter();
			sw.write(o);
			byte[] b = sw.getBytes();
            log.finest("Length: " + b.length);
            if(log.isLoggable(Level.FINEST)) {
                log.finest("Bytes:" + LogHelper.byteArrayToString(b));
            }
			String sql = "INSERT INTO " + o.getClass().getSimpleName() + " VALUES (" + PARAMS.substring(0, 2*keys.length + 1) + ")";
			pstmt = con.prepareStatement(sql);
			for(int i = 0; i < keys.length; i++) {
			    setKey(pstmt, keys, i, i+1, o);
			}
			pstmt.setBytes(keys.length+1, b);
			int n = pstmt.executeUpdate();
			if(n != 1) {
                log.fine("Insert not OK");
			    throw new IsamException("Insert failed");
			} else {
			    log.fine("Insert OK");
			}
		} catch (SQLException e) {
			throw new IsamException("Database problem", e);
		} catch (IllegalAccessException e) {
			throw new IsamException("Cannot access key field", e);
		} finally {
			try {
				if(pstmt != null) pstmt.close();
			} catch (SQLException e) {
				// not much that can be done
			}
		}
	}
    public <T> void update(T o) throws IsamException, RecordException {
        log.fine("Update class " + o.getClass().getName() + " with value " + o);
        Class<?> t = o.getClass();
        KeyInfo[] keys = KeyInfoArrayCache.analyze(t);
        PreparedStatement pstmt = null;
        try {
            StructWriter sw = new StructWriter();
            sw.write(o);
            byte[] b = sw.getBytes();
            log.finest("Length: " + b.length);
            if(log.isLoggable(Level.FINEST)) {
                log.finest("Bytes:" + LogHelper.byteArrayToString(b));
            }
            String sql = "UPDATE " + o.getClass().getSimpleName() + " SET recblob = ? WHERE " + keys[0].getField().getName() + " = ?";
            pstmt = con.prepareStatement(sql);
            setKey(pstmt, keys, 0, 2, o);
            pstmt.setBytes(1, b);
            int n = pstmt.executeUpdate();
            if(n < 1) {
                log.fine("Updated not OK");
                throw new IsamException("Update failed - key not found");
            } else {
                log.fine("Update OK");
            }
        } catch (SQLException e) {
            throw new IsamException("Database problem", e);
        } catch (IllegalAccessException e) {
            throw new IsamException("Cannot access key field", e);
        } finally {
            try {
                if(pstmt != null) pstmt.close();
            } catch (SQLException e) {
                // not much that can be done
            }
        }
    }
    public <T, TK> void delete(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("Delete class " + t.getName() + " with key " + k);
        KeyInfo[] keys = KeyInfoArrayCache.analyze(t);
        PreparedStatement pstmt = null;
        ResultSet rs = null;
        try {
            pstmt = con.prepareStatement("DELETE FROM " + t.getSimpleName() + " WHERE " + keys[k.getIx()].getField().getName() + " = ?");
            setKey(pstmt, keys, k);
            int n = pstmt.executeUpdate();
            if(n < 1) {
                log.fine("Delete not OK");
                throw new IsamException("Delete failed - key not found");
            } else {
                log.fine("Delete OK");
            }
        } catch (SQLException e) {
            throw new IsamException("Database problem", e);
        } finally {
            try {
                if(rs != null) rs.close();
                if(pstmt != null) pstmt.close();
            } catch (SQLException e) {
                // not much that can be done
            }
        }
    }
	public void close() throws IsamException {
		try {
			con.close();
			log.fine("Closed");
		} catch (SQLException e) {
			throw new IsamException("Database problem", e);
		}
	}
	public <T, TK> IsamResult<T> readGE(Class<T> t, Key<TK> k) throws IsamException {
        log.fine("ReadGE class " + t.getName() + " with key " + k);
		return readXX(t, k, ">=", "ASC");
	}
	public <T, TK> IsamResult<T> readGT(Class<T> t, Key<TK> k) throws IsamException {
        log.fine("ReadGT class " + t.getName() + " with key " + k);
		return readXX(t, k, ">", "ASC");
	}
	public <T, TK> IsamResult<T> readLE(Class<T> t, Key<TK> k) throws IsamException {
        log.fine("ReadLE class " + t.getName() + " with key " + k);
		return readXX(t, k, "<=", "DESC");
	}
	public <T, TK> IsamResult<T> readLT(Class<T> t, Key<TK> k) throws IsamException {
        log.fine("ReadLT class " + t.getName() + " with key " + k);
		return readXX(t, k, "<", "DESC");
	}
	private <TK> void setKey(PreparedStatement pstmt, KeyInfo[] keys, Key<TK> k) throws SQLException, IsamException {
		FieldType ft = keys[k.getIx()].getTyp();
		switch(ft) {
			case INT1:
				pstmt.setByte(1, (Byte)k.getValue());
				break;
			case INT2:
				pstmt.setShort(1, (Short)k.getValue());
				break;
            case INT4:
                pstmt.setInt(1, (Integer)k.getValue());
                break;
            case INT8:
                pstmt.setLong(1, (Long)k.getValue());
                break;
            case FIXSTR:
            case FIXSTRNULTERM:
            case VARSTR:
            case VARFIXSTR:
				pstmt.setString(1, (String)k.getValue());
				break;
			default:
				throw new IsamException(ft.toString() + " in " + k.getClass().getName() + " can not be used as a key");
		}
		return;
	}
    private <T> void setKey(PreparedStatement stmt, KeyInfo[] keys, int i, int param, T o) throws IllegalArgumentException, SQLException, IllegalAccessException, IsamException {
        FieldType ft = keys[i].getTyp();
        switch(ft) {
            case INT1:
                stmt.setByte(param, keys[i].getField().getByte(o));
                break;
            case INT2:
                stmt.setShort(param, keys[i].getField().getShort(o));
                break;
            case INT4:
                stmt.setInt(param, keys[i].getField().getInt(o));
                break;
            case INT8:
                stmt.setLong(param, keys[i].getField().getLong(o));
                break;
            case FIXSTR:
            case FIXSTRNULTERM:
            case VARSTR:
            case VARFIXSTR:
                stmt.setString(param, (String)keys[i].getField().get(o));
                break;
            default:
                throw new IsamException(ft.toString() + " in " + o.getClass().getName() + " can not be used as a key");
        }
    }
	private <T, TK> IsamResult<T> readXX(Class<T> t, Key<TK> k, String comp, String sort) throws IsamException {
		KeyInfo[] keys = KeyInfoArrayCache.analyze(t);
		PreparedStatement pstmt = null;
		ResultSet rs = null;
		try {
            Connection con2 = DriverManager.getConnection(conurl, un, pw);
			pstmt = con2.prepareStatement("SELECT recblob FROM " + t.getSimpleName() + " WHERE " + keys[k.getIx()].getField().getName() + " " + comp + " ? ORDER BY " + keys[k.getIx()].getField().getName() + " " + sort);
			setKey(pstmt, keys, k);
			rs = pstmt.executeQuery();
	        return new EmulatedIsamResult<T>(t, con2, pstmt, rs);
		} catch (SQLException e) {
			try {
				if(rs != null) rs.close();
				if(pstmt != null) pstmt.close();
			} catch (SQLException e2) {
				// not much that can be done
			}
			throw new IsamException("Database problem", e);
		}
	}
}
