/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.local;

import dk.vajhoej.isam.IsamException;

/**
 * Interface NativeIsamt represents the native implementation of the local ISAM source.
 */
public interface NativeIsam {
    /**
     * Greater than or equal. 
     */
    public static final int COMP_GE = 1;
    /**
     * Greater than. 
     */
    public static final int COMP_GT = 2;
    /**
     * Less than or equal. 
     */
    public static final int COMP_LE = 3;
    /**
     * Less than. 
     */
    public static final int COMP_LT = 4;
    /**
     * Open isam file.
     * @param fnm file name
     * @param readonly readonly mode
     * @throws IsamException if problem
     */
    public void open(String fnm, boolean readonly) throws IsamException;
    /**
     * Read record by key.
     * @param ix key index
     * @param k key value
     * @param buf buffer to receive data
     * @return status (less than or equal zero => error)
     * @throws IsamException if problem
     */
    public int read(int ix, Object k, byte[] buf) throws IsamException;
    /**
     * Create new record.
     * @param buf buffer with data
     * @return status (less than or equal zero => error)
     * @throws IsamException
     */
    public int create(byte[] buf) throws IsamException;
    /**
     * Update existing record.
     * @param buf buffer with data
     * @return status (less than or equal zero => error)
     * @throws IsamException if problem
     */
    public int update(byte[] buf) throws IsamException;
    /**
     * Update existing record and do locking first.
     * @param ix key index
     * @param k key value
     * @param buf buffer with data
     * @return status (less than or equal zero => error)
     * @throws IsamException if problem
     */
    public int update(byte[] buf, int ix, Object k) throws IsamException;
    /**
     * Delete record by key.
     * @param ix key index
     * @param k key value
     * @return status (less than or equal zero => error)
     * @throws IsamException if problem
     */
    public int delete(int ix, Object k) throws IsamException;
    /**
     * Search for record by key.
     * @param ix key index
     * @param k key value
     * @param comp key comparison
     * @param buf buffer to receive data
     * @return status (less than or equal zero => error)
     * @throws IsamException if problem
     */
    public int find(int ix, Object k, int comp, byte[] buf) throws IsamException;
    /**
     * Next/previous record (GE and GT => next, LE and LT => previous).
     * @param buf buffer to receive data
     * @return status (less than or equal zero => error)
     * @throws IsamException if problem
     */
    public int findnext(byte[] buf) throws IsamException;
    /**
     * Close.
     * @throws IsamException if problem
     */
    public void close() throws IsamException;
    /**
     * Does this implementation support reverse find for LE and LT.
     * @return true or false
     */
    public boolean supportReverse();
    /**
     * Does this implementation support modification of records without lock.
     * @return true or false
     */
    public boolean supportModWithoutLock();
}
