/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.remote;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamResult;
import dk.vajhoej.isam.Key;
import dk.vajhoej.record.RecordException;

/**
 * Class ObjectClient implement IsamProtocolClient using ObjectStream's.
 */
public class ObjectClient implements IsamProtocolClient {
    private static final boolean USEBUFFER = true;
    private static Logger log = Logger.getLogger(ObjectClient.class.getName());
    private InputStream is;
    private OutputStream os;
    private ObjectInputStream ois;
    private ObjectOutputStream oos;
    /**
     * Create instance of ObjectClient.
     * @param is input stream to wrap in input object stream
     * @param os output stream to wrap in output object stream
     */
    public ObjectClient(InputStream is, OutputStream os) {
        if(USEBUFFER) {
            this.is = new BufferedInputStream(is);
            this.os = new BufferedOutputStream(os);
        } else {
            this.is = is;
            this.os = os;
        }
        log.finest("Streams open");
    }
    public void handshake() throws IsamException {
        try {
            os.write(IsamProtocol.O_V1);
            os.flush();
            // note that the order and the flush are necessary to avoid hanging
            ois = new ObjectInputStream(is);
            oos = new ObjectOutputStream(os);
            oos.flush();
            log.finer("Handshake complete");
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    @SuppressWarnings("unchecked")
    public <T, TK> T read(Class<T> t, Key<TK> k) throws IsamException {
        log.fine("Read class " + t.getName() + " with key " + k);
        try {
            oos.writeObject(IsamProtocol.READ);
            oos.reset();
            oos.writeObject(t);
            oos.writeObject(k);
            oos.flush();
            log.finer("Command send to server");
            int ok = (Integer)ois.readObject();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", (IsamException)ois.readObject());
            } else {
                log.fine("OK");
            }
            T res = (T)ois.readObject();
            if(res != null) {
                log.finest("Read " + res.getClass().getName() + " object");
            }
            return res;
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public <T> void create(T o) throws IsamException, RecordException {
        log.fine("Create class " + o.getClass().getName() + " with value " + o);
        try {
            oos.writeObject(IsamProtocol.CREATE);
            oos.reset();
            oos.writeObject(o);
            oos.flush();
            log.finer("Command send to server");
            int ok = (Integer)ois.readObject();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", (IsamException)ois.readObject());
            } else {
                log.fine("OK");
            }
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public <T> void update(T o) throws IsamException, RecordException {
        log.fine("Update class " + o.getClass().getName() + " with value " + o);
        try {
            oos.writeObject(IsamProtocol.UPDATE);
            oos.reset();
            oos.writeObject(o);
            oos.flush();
            log.finer("Command send to server");
            int ok = (Integer)ois.readObject();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", (IsamException)ois.readObject());
            } else {
                log.fine("OK");
            }
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public <T, TK> void delete(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("Delete class " + t.getName() + " with key " + k);
        try {
            oos.writeObject(IsamProtocol.DELETE);
            oos.reset();
            oos.writeObject(t);
            oos.writeObject(k);
            oos.flush();
            log.finer("Command send to server");
            int ok = (Integer)ois.readObject();
            if(ok != IsamProtocol.OK) {
                log.fine("OK");
                throw new IsamException("Remote server exception", (IsamException)ois.readObject());
            } else {
                log.fine("OK");
            }
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public void close() throws IsamException {
        log.fine("Close");
        try {
            oos.writeObject(IsamProtocol.CLOSE);
            oos.flush();
            log.finer("Command send to server");
            int ok = (Integer)ois.readObject();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", (IsamException)ois.readObject());
            } else {
                log.fine("OK");
            }
            log.finest("Closing streams");
            oos.close();
            ois.close();
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public <T, TK> IsamResult<T> readXX(int cmd, Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("ReadXX class " + t.getName() + " with key " + k);
        try {
            oos.writeObject(cmd);
            oos.reset();
            oos.writeObject(t);
            oos.writeObject(k);
            oos.flush();
            log.finer("Command send to server");
            int ok = (Integer)ois.readObject();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", (IsamException)ois.readObject());
            } else {
                log.fine("OK");
            }
            return new RemoteIsamResult<T>(this);
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public boolean resultRead() throws IsamException, RecordException {
        log.fine("Result read");
        try {
            oos.writeObject(IsamProtocol.RESULT_READ);
            oos.flush();
            log.finer("Command send to server");
            int ok = (Integer)ois.readObject();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", (IsamException)ois.readObject());
            } else {
                log.fine("OK");
            }
            return (Boolean)ois.readObject();
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    @SuppressWarnings("unchecked")
    public <T> T resultCurrent() throws IsamException, RecordException {
        log.fine("Result current");
        try {
            oos.writeObject(IsamProtocol.RESULT_CURRENT);
            oos.flush();
            log.finer("Command send to server");
            int ok = (Integer)ois.readObject();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", (IsamException)ois.readObject());
            } else {
                log.fine("OK");
            }
            T res = (T)ois.readObject();
            log.finest("Read " + res.getClass().getName() + " object");
            return res;
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public void resultClose() throws IsamException {
        log.fine("Result close");
        try {
            oos.writeObject(IsamProtocol.RESULT_CLOSE);
            oos.flush();
            log.finer("Command send to server");
            int ok = (Integer)ois.readObject();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", (IsamException)ois.readObject());
            } else {
                log.fine("OK");
            }
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
}
