/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.rms;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Parse Pascal source code and generate Java bean classes for types.
 * Run as:
 * <pre>
 *   java dk.vajhoej.vms.rms.PascalParse <pascal file>
 * </pre>
 */
public class PascalParse {
    private static final int MISSING = -1;
    private static Map<String, String> pascal2java = new HashMap<String, String>(); 
    private static Map<String, String> pascal2struct = new HashMap<String, String>();
    private static Map<String, Integer> pascal2length = new HashMap<String, Integer>();
    private static void map(String pascal, String java, String struct, int length) {
        pascal2java.put(pascal, java);
        pascal2struct.put(pascal, struct);
        pascal2length.put(pascal, length);
    }
    static {
        map("integer", "int", "INT4", MISSING);
        map("unsigned", "long", "UINT4", MISSING);
        map("real", "double", "FP4", MISSING);
        map("single", "double", "FP4", MISSING);
        map("double", "double", "FP8", MISSING);
        map("boolean", "boolean", "BOOLEAN", 4);
    }
    private static class Field {
        private String name;
        private String javaType;
        private String structType;
        private int length;
        private int key;
        public Field(String name, String type, int key) {
            this.name = name.toLowerCase();
            this.javaType = pascal2java.get(type);
            this.structType = pascal2struct.get(type);
            this.length = pascal2length.get(type);
            this.key = key;
        }
        public String getName() {
            return name;
        }
        public String getJavaType() {
            return javaType;
        }
        public String getStructType() {
            return structType;
        }
        public int getLength() {
            return length;
        }
        public int getKey() {
            return key;
        }
    }
    private static String capitalize(String s) {
        return s.substring(0, 1).toUpperCase() + s.substring(1);
    }
    private static String stripComments(String s) {
        return s.replaceAll("\\{.*?\\}", "").replaceAll("\\(\\*.*?\\*\\)", "");
    }
    private static void genClass(String recnam, List<Field> fld) {
        System.out.println("import java.io.Serializable;");
        System.out.println();
        System.out.println("import dk.vajhoej.isam.KeyField;");
        System.out.println("import dk.vajhoej.record.Alignment;");
        System.out.println("import dk.vajhoej.record.FieldType;");
        System.out.println("import dk.vajhoej.record.Struct;");
        System.out.println("import dk.vajhoej.record.StructField;");
        System.out.println();
        System.out.println("@Struct(alignment=Alignment.NATURAL,endpad=true)");
        System.out.println("public class " + recnam + " implements Serializable {");
        System.out.println("    private static final long serialVersionUID = 1L;");
        int nfld = 0;
        for(Field fnt : fld) {
            if(fnt.getKey() > MISSING) {
                System.out.println("    @KeyField(n=" + fnt.getKey() + ")");
            }
            if(fnt.getLength() > MISSING) {
                System.out.println("    @StructField(n=" + nfld + ",type=FieldType." + fnt.getStructType() + ",length=" + fnt.getLength() + ")");
            } else {
                System.out.println("    @StructField(n=" + nfld + ",type=FieldType." + fnt.getStructType() + ")");
            }
            System.out.println("    private " + fnt.getJavaType() + " " + fnt.getName() + ";");
            nfld++;
        }
        for(Field fnt : fld) {
            System.out.println("    public " + fnt.getJavaType() + " get" + capitalize(fnt.getName()) + "() {");
            System.out.println("        return " + fnt.getName() + ";");
            System.out.println("    }");
            System.out.println("    public void set" + capitalize(fnt.getName()) + "(" + fnt.getJavaType() + " " + fnt.getName() + ") {");
            System.out.println("        this." + fnt.getName() + " = " + fnt.getName() + ";");
            System.out.println("    }");
        }
        System.out.println("}");
        System.out.println();
    }
    private static Pattern varchrpat = Pattern.compile("(\\w+)(?:\\s*=\\s*varying\\s*\\[)(\\d+)(?:\\] of char;)", Pattern.CASE_INSENSITIVE + Pattern.DOTALL);
    private static Pattern pckarrchrpat = Pattern.compile("(\\w+)(?:\\s*=\\s*packed array\\s*\\[1\\.\\.)(\\d+)(?:\\] of char;)", Pattern.CASE_INSENSITIVE + Pattern.DOTALL);
    private static void processStrings(String s) {
        Matcher varchr = varchrpat.matcher(s);
        while(varchr.find()) {
            map(varchr.group(1), "String", "VARFIXSTR", Integer.parseInt(varchr.group(2)));
        }
        Matcher pckarrchr = pckarrchrpat.matcher(s);
        while(pckarrchr.find()) {
            map(pckarrchr.group(1), "String", "FIXSTR", Integer.parseInt(pckarrchr.group(2)));
        }
    }
    private static Pattern wordpat = Pattern.compile("(\\w+)(?:\\s*=\\s*)(\\[WORD\\]\\s*-32768\\.\\.32767;)", Pattern.CASE_INSENSITIVE + Pattern.DOTALL);
    private static Pattern uwordpat = Pattern.compile("(\\w+)(?:\\s*=\\s*)(\\[WORD\\]\\s*0\\.\\.65535;)", Pattern.CASE_INSENSITIVE + Pattern.DOTALL);
    private static Pattern bytepat = Pattern.compile("(\\w+)(?:\\s*=\\s*)(\\[BYTE\\]\\s*-128\\.\\.127;)", Pattern.CASE_INSENSITIVE + Pattern.DOTALL);
    private static Pattern ubytepat = Pattern.compile("(\\w+)(?:\\s*=\\s*)(\\[BYTE\\]\\s*0\\.\\.255;)", Pattern.CASE_INSENSITIVE + Pattern.DOTALL);
    private static void processWordByte(String s) {
        Matcher tword = wordpat.matcher(s);
        while(tword.find()) {
            map(tword.group(1), "short", "INT2", MISSING);
        }
        Matcher tuword = uwordpat.matcher(s);
        while(tuword.find()) {
            map(tuword.group(1), "int", "UINT2", MISSING);
        }
        Matcher tbyte = bytepat.matcher(s);
        while(tbyte.find()) {
            map(tbyte.group(1), "byte", "INT1", MISSING);
        }
        Matcher tubyte = ubytepat.matcher(s);
        while(tubyte.find()) {
            map(tubyte.group(1), "short", "UINT1", MISSING);
        }
    }
    private static Pattern fldpat = Pattern.compile("(\\w+)(?:\\s*:\\s*)((?:\\[KEY\\()(\\d+)(?:\\)\\])|())(?:\\s*)(\\w+)(?:\\s*;)");
    private static void processRecord(String recnam, String s) {
        List<Field> fld = new ArrayList<Field>();
        Matcher m = fldpat.matcher(s);
        while(m.find()) {
            fld.add(new Field(m.group(1), m.group(5), m.group(3) != null ? Integer.parseInt(m.group(3)) : MISSING));
        }
        genClass(capitalize(recnam.toLowerCase()), fld);
    }
    private static Pattern recpat = Pattern.compile("(\\w+)(?:\\s*=\\s*record\\s+)(.*?)(?:\\s+end\\s*;)", Pattern.CASE_INSENSITIVE + Pattern.DOTALL);
    /**
     * Process a file.
     * @param s content of file
     */
    public static void processFile(String s) {
        processStrings(s);
        processWordByte(s);
        Matcher m = recpat.matcher(stripComments(s));
        while(m.find()) {
            processRecord(m.group(1), m.group(2));
            map(m.group(1), capitalize(m.group(1)), "STRUCT", MISSING);
        }
    }
    /**
     * PascalParse main.
     * @param args command line arguments - first and only argument is name of pascal file 
     * @throws IOException if problem with file
     */
    public static void main(String[] args) {
        try {
            BufferedReader br = new BufferedReader(new FileReader(args[0]));
            StringBuilder sb = new StringBuilder();
            String line;
            while((line = br.readLine()) != null) {
                sb.append(line);
                sb.append(System.getProperty("line.separator"));
            }
            br.close();
            processFile(sb.toString());
        } catch(Exception ex) {
            ex.printStackTrace();
        }
    }
}
